from unittest import TestCase

import numpy as np

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestIntersect(TestCase):
    def setUp(self) -> None:
        # inputs
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]
        self.stack = Raster(self.predictors)

        # test results
        self.result = None

    def tearDown(self) -> None:
        self.stack.close()
        self.result.close()

    def test_intersect_defaults(self):
        self.result = self.stack.intersect()

        # check raster object
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, self.stack.count)
        self.assertEqual(self.result.read(masked=True).count(), 810552)

        # test nodata value is recognized
        self.assertEqual(self.result.read(masked=True).min(), 1.0)
        self.assertEqual(self.result.read(masked=True).max(), 255.0)

    def test_intersect_custom_dtype(self):
        self.result = self.stack.intersect(dtype=np.int16)

        # check raster object
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, self.stack.count)
        self.assertEqual(self.result.read(masked=True).count(), 810552)

        # test nodata value is recognized
        self.assertEqual(self.result.read(masked=True).min(), 1)
        self.assertEqual(self.result.read(masked=True).max(), 255)

    def test_intersect_custom_nodata(self):
        self.result = self.stack.intersect(dtype=np.int16, nodata=-999)

        # check raster object
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, self.stack.count)
        self.assertEqual(self.result.read(masked=True).count(), 810552)

        # test nodata value is recognized
        self.assertEqual(self.result.read(masked=True).min(), 1)
        self.assertEqual(self.result.read(masked=True).max(), 255)

    def test_intersect_in_memory(self):
        self.result = self.stack.intersect(in_memory=True)

        # check raster object
        self.assertIsInstance(self.result, Raster)



from unittest import TestCase

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster
import geopandas as gpd
from sklearn.preprocessing import StandardScaler


class TestAlter(TestCase):
    def setUp(self) -> None:
        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.stack = Raster(predictors)
        points = gpd.read_file(nc.points)
        data = self.stack.extract_vector(points)
        self.data = data.dropna()

    def tearDown(self) -> None:
        self.stack.close()

    def test_alter(self):
        scaler = StandardScaler()
        scaler.fit(self.data.drop(columns=["geometry"]).values)
        out = self.stack.alter(scaler)

        self.assertIsInstance(out, Raster)
        self.assertEqual(out.shape, self.stack.shape)



from unittest import TestCase

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster
import rasterio
import numpy as np
import os
from tempfile import NamedTemporaryFile


class TestRename(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]

    def test_rename_inplace(self):
        stack = Raster(self.predictors)
        band3_stats = stack.lsat7_2000_30.mean()

        # rename band 3
        stack.rename(names={"lsat7_2000_30": "new_name"}, in_place=True)

        # check that renaming occurred in Raster
        self.assertEqual(list(stack.names)[2], "new_name")
        self.assertNotIn("lsat7_2000_30", stack.names)

        # check that Raster layer properties also renamed
        self.assertIn("new_name", dir(stack))
        self.assertNotIn("lsat7_2000_30", dir(stack))

        # check that internal name of RasterLayer was also renamed
        self.assertEqual(stack.iloc[2].name, "new_name")

        # check that the RasterLayer attached to the new name is the same
        self.assertEqual(stack["new_name"].mean(), band3_stats)
        self.assertEqual(stack.new_name.mean(), band3_stats)
        self.assertEqual(stack.iloc[2].mean(), band3_stats)

        # check that a new Raster object derived from the renamed data
        # have the right names
        new_raster = Raster(src=stack.iloc[2])
        self.assertIn("new_name", new_raster.names)

    def test_rename_with_copy(self):
        stack = Raster(self.predictors)
        names = list(stack.names)
        band3_stats = stack.lsat7_2000_30.mean()

        # rename band 3
        result = stack.rename(names={"lsat7_2000_30": "new_name"},
                              in_place=False)

        # check that original is untouched
        self.assertEqual(list(stack.names), names)

        # check that renaming occurred in Raster
        self.assertEqual(list(result.names)[2], "new_name")
        self.assertNotIn("lsat7_2000_30", result.names)

        # check that Raster layer properties also renamed
        self.assertIn("new_name", dir(result))
        self.assertNotIn("lsat7_2000_30", dir(result))

        # check that internal name of RasterLayer was also renamed
        self.assertEqual(result.iloc[2].name, "new_name")

        # check that the RasterLayer attached to the new name is the same
        self.assertEqual(result["new_name"].mean(), band3_stats)
        self.assertEqual(result["new_name"].mean(), band3_stats)
        self.assertEqual(result.new_name.mean(), band3_stats)
        self.assertEqual(result.iloc[2].mean(), band3_stats)

        # check that a new Raster object derived from the renamed data
        # have the right names
        new_raster = Raster(src=result.iloc[2])
        self.assertIn("new_name", new_raster.names)

    def rename_multiband(self):
        # Create a fake 3-band image for testing
        arr = np.random.rand(3, 64, 64)
        file = NamedTemporaryFile(prefix="test", suffix=".tif").name
        layer_name = os.path.basename(file).split(".")[0]
        layer_names = ["_".join([layer_name, str(i)]) for i in [1, 2 ,3]]

        with rasterio.open(file, "w", width=64, height=64, count=3, dtype=np.float32) as dst:
            dst.write(arr)

        r = Raster(file)
        self.assertListEqual(list(r.names), layer_names)

        renamed = r.rename(dict(zip(r.names, ["Red", "Green", "Blue"])))
        self.assertListEqual(list(renamed.names), ["Red", "Green", "Blue"])

    def rename_in_memory(self):
        # Create a fake 3-band image for testing
        arr = np.random.rand(3, 64, 64)
        file = NamedTemporaryFile(prefix="test", suffix=".tif").name

        with rasterio.open(file, "w", width=64, height=64, count=3, dtype=np.float32) as dst:
            dst.write(arr)

        r = Raster(file)
        in_memory = r.aggregate((32, 32), in_memory=True)

        renamed = r.rename(dict(zip(in_memory.names, ["Red", "Green", "Blue"])))
        self.assertListEqual(list(renamed.names), ["Red", "Green", "Blue"])



import numpy as np
import pandas as pd


def stack_constants(flat_pixels, constants, names=None):
    """Column stack any constant values into the flat_pixels array.

    Used to add additional constant features to the Raster object.

    Parameters
    ----------
    flat_pixels : ndarray
        2d numpy array representing the flattened raster data in
        (sample_n, band_values) format.

    constants : list-like object, 1d array, or dict
        Array of constant values to be added to the flat_pixels array
        as additional features.

        If a dict is passed, the dict keys must refer to names of the
        features in the flat_pixels array, and the values will replace
        these features with constant values.

    names : list-like object (optional, default=None)
        Names of the raster layers.
    """
    if isinstance(constants, (int, float)):
        constants = [constants]

    if isinstance(constants, list):
        constants = np.asarray(constants)
        constants = np.broadcast_to(
            constants, (flat_pixels.shape[0], constants.shape[0])
        )
        flat_pixels = np.column_stack((flat_pixels, constants))

    elif isinstance(constants, dict):

        keys_not_in_raster = [i for i in constants.keys() if i not in names]

        if len(keys_not_in_raster) > 0:
            raise ValueError(
                "The following keys are not in the raster: {x}".format(
                    x=keys_not_in_raster
                )
            )

        flat_pixels = pd.DataFrame(flat_pixels, columns=names)

        for key, value in constants.items():
            flat_pixels[key] = value

        flat_pixels = flat_pixels.values

    elif isinstance(constants, np.ndarray):
        raise ValueError("constants must be a list or a numpy.ndarray")

    return flat_pixels


def predict_output(img, estimator, constants=None, names=None):
    """Prediction function for classification or regression response.

    Parameters
    ----
    img : tuple (window, numpy.ndarray)
        A window object, and a 3d ndarray of raster data with the
        dimensions in order of (band, rows, columns).

    estimator : estimator object implementing 'fit'
        The object to use to fit the data.

    constants : list-like object, 1d array, or dict
        Array of constant values to be added to the flat_pixels array
        as additional features.

        If a dict is passed, the dict keys must refer to names of the
        features in the flat_pixels array, and the values will replace
        these features with constant values.

    names : list-like object (optional, default=None)
        Names of the raster layers.

    Returns
    -------
    numpy.ndarray
        2d numpy array representing a single band raster containing the
        classification or regression result.
    """
    window, img = img
    img = np.ma.masked_invalid(img)

    # reorder into rows, cols, bands(transpose)
    n_features, rows, cols = img.shape[0], img.shape[1], img.shape[2]

    # reshape into 2D array (rows=sample_n, cols=band_values)
    n_samples = rows * cols
    flat_pixels = img.transpose(1, 2, 0).reshape((n_samples, n_features))

    # create mask for NaN values
    flat_pixels_mask = flat_pixels.mask.copy()

    # fill nans for prediction
    flat_pixels = flat_pixels.filled(0)

    # add constants
    if constants is not None:
        flat_pixels = stack_constants(flat_pixels, constants, names)

    # predict and replace mask
    result = estimator.predict(flat_pixels)
    result = np.ma.masked_array(data=result, mask=flat_pixels_mask.any(axis=1))

    # reshape the prediction from a 1D into 3D array [band, row, col]
    result = result.reshape((1, window.height, window.width))

    return result


def predict_prob(img, estimator, constants=None, names=None):
    """Class probabilities function.

    Parameters
    ----------
    img : tuple (window, numpy.ndarray)
        A window object, and a 3d ndarray of raster data with the
        dimensions in order of (band, rows, columns).

    estimator : estimator object implementing 'fit'
        The object to use to fit the data.

    constants : list-like object, 1d array, or dict
        Array of constant values to be added to the flat_pixels array
        as additional features.

        If a dict is passed, the dict keys must refer to names of the
        features in the flat_pixels array, and the values will replace
        these features with constant values.

    names : list-like object (optional, default=None)
        Names of the raster layers.

    Returns
    -------
    numpy.ndarray
        Multi band raster as a 3d numpy array containing the
        probabilities associated with each class. ndarray dimensions
        are in the order of (class, row, column).
    """
    window, img = img

    # reorder into rows, cols, bands (transpose)
    n_features, rows, cols = img.shape[0], img.shape[1], img.shape[2]
    img = np.ma.masked_invalid(img)
    mask2d = img.mask.any(axis=0)

    # then resample into 2D array (rows=sample_n, cols=band_values)
    n_samples = rows * cols
    flat_pixels = img.transpose(1, 2, 0).reshape((n_samples, n_features))

    # fill mask with zeros for prediction
    flat_pixels = flat_pixels.filled(0)

    # add constants
    if constants is not None:
        flat_pixels = stack_constants(flat_pixels, constants, names)

    # predict probabilities
    result_proba = estimator.predict_proba(flat_pixels)

    # reshape class probabilities back to 3D array [class, rows, cols]
    result_proba = result_proba.reshape(
        (window.height, window.width, result_proba.shape[1])
    )

    # reshape band into rasterio format [band, row, col]
    result_proba = result_proba.transpose(2, 0, 1)

    # repeat mask for n_bands
    mask3d = np.repeat(
        a=mask2d[np.newaxis, :, :], repeats=result_proba.shape[0], axis=0
    )

    # convert proba to masked array
    result_proba = np.ma.masked_array(result_proba, mask=mask3d, fill_value=np.nan)

    return result_proba


def predict_multioutput(img, estimator, constants=None, names=None):
    """Multi-target prediction function.

    Parameters
    ----------
    img : tuple (window, numpy.ndarray)
        A window object, and a 3d ndarray of raster data with the
        dimensions in order of (band, rows, columns).

    estimator : estimator object implementing 'fit'
        The object to use to fit the data.

    constants : list-like object, 1d array, or dict
        Array of constant values to be added to the flat_pixels array
        as additional features.

        If a dict is passed, the dict keys must refer to names of the
        features in the flat_pixels array, and the values will replace
        these features with constant values.

    names : list-like object (optional, default=None)
        Names of the raster layers.

    Returns
    -------
    numpy.ndarray
        3d numpy array representing the multi-target prediction result
        with the dimensions in the order of (target, row, column).
    """
    window, img = img

    # reorder into rows, cols, bands(transpose)
    n_features, rows, cols = img.shape[0], img.shape[1], img.shape[2]
    img = np.ma.masked_invalid(img)
    mask2d = img.mask.any(axis=0)

    # reshape into 2D array (rows=sample_n, cols=band_values)
    n_samples = rows * cols
    flat_pixels = img.transpose(1, 2, 0).reshape((n_samples, n_features))
    flat_pixels = flat_pixels.filled(0)

    # add constants
    if constants is not None:
        flat_pixels = stack_constants(flat_pixels, constants, names)

    # predict probabilities
    result = estimator.predict(flat_pixels)

    # reshape class probabilities back to 3D array [class, rows, cols]
    result = result.reshape((window.height, window.width, result.shape[1]))

    # reshape band into rasterio format [band, row, col]
    result = result.transpose(2, 0, 1)

    # repeat mask for n_bands
    mask3d = np.repeat(a=mask2d[np.newaxis, :, :], repeats=result.shape[0], axis=0)

    # convert proba to masked array
    result = np.ma.masked_array(result, mask=mask3d, fill_value=np.nan)

    return result



from unittest import TestCase

import numpy as np

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestSample(TestCase):
    def setUp(self) -> None:
        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.stack = Raster(predictors)
        self.strata = Raster(nc.strata)
    
    def tearDown(self) -> None:
        self.stack.close()
        self.strata.close()

    def test_sample_strata(self):
        # extract using a strata raster and returning two arrays
        size = 100
        categories = self.strata.read(masked=True).flatten()
        categories = categories[~categories.mask]
        n_categories = np.unique(categories).shape[0]
        n_samples = size * n_categories

        X, xy = self.stack.sample(size=size, strata=self.strata, return_array=True)
        self.assertEqual(X.shape, (n_samples, 6))
        self.assertEqual(xy.shape, (n_samples, 2))

        # extract using a strata raster and returning a dataframe
        samples = self.stack.sample(size=size, strata=self.strata, return_array=False)
        self.assertEqual(samples.shape, (n_samples, 7))
    
    def test_sample_no_strata(self):
        size = 100
        X, xy = self.stack.sample(size=size, return_array=True)
        self.assertEqual(X.shape, (size, 6))
        self.assertEqual(xy.shape, (size, 2))

        samples = self.stack.sample(size=size, return_array=False)
        self.assertEqual(samples.shape, (size, 7))


import os

band1 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_10.tif')
band2 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_20.tif')
band3 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_30.tif')
band4 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_40.tif')
band5 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_50.tif')
band7 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_70.tif')
strata = os.path.join(os.path.dirname(__file__), 'strata.tif')
multiband = os.path.join(os.path.dirname(__file__), 'landsat_multiband.tif')
labelled_pixels = os.path.join(os.path.dirname(__file__), 'landsat96_labelled_pixels.tif')
points = os.path.join(os.path.dirname(__file__), 'landsat96_points.shp')
polygons = os.path.join(os.path.dirname(__file__), 'landsat96_polygons.shp')
extracted_pixels = os.path.join(os.path.dirname(__file__), 'extracted_pixels.txt')



import numpy as np
from abc import ABC


class RasterStats(ABC):
    def _stats(self, max_pixels):
        rel_width = self.shape[1] / max_pixels

        if rel_width > 1:
            col_scaling = round(max_pixels / rel_width)
            row_scaling = max_pixels - col_scaling
        else:
            col_scaling = round(max_pixels * rel_width)
            row_scaling = max_pixels - col_scaling

        out_shape = (row_scaling, col_scaling)
        arr = self.read(masked=True, out_shape=out_shape)
        return arr.reshape((arr.shape[0], arr.shape[1] * arr.shape[2]))

    def min(self, max_pixels=10000):
        arr = self._stats(max_pixels)
        return np.nanmin(arr, axis=1).data

    def max(self, max_pixels=10000):
        arr = self._stats(max_pixels)
        return np.nanmax(arr, axis=1).data

    def mean(self, max_pixels=10000):
        arr = self._stats(max_pixels)
        return np.nanmean(arr, axis=1).data

    def median(self, max_pixels=10000):
        arr = self._stats(max_pixels)
        return np.nanmedian(arr, axis=1).data

    def stddev(self, max_pixels=10000):
        arr = self._stats(max_pixels)
        return np.nanstd(arr, axis=1).data



import math

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.ticker as mticker


def discrete_cmap(N, base_cmap=None):
    """Create an N-bin discrete colormap from the specified input map.

    Source:
    https://gist.github.com/jakevdp/91077b0cae40f8f8244a

    Parameters
    ----------
    N : int
        The number of colors in the colormap

    base_cmap : str
        The name of the matplotlib cmap to convert into a discrete map.

    Returns
    -------
    matplotlib.cmap
        The cmap converted to a discrete map.
    """

    base = plt.cm.get_cmap(base_cmap)
    color_list = base(np.linspace(0, 1, N))
    cmap_name = base.name + str(N)

    return base.from_list(cmap_name, color_list, N)


class RasterPlot(object):
    def plot(
        self,
        cmap=None,
        norm=None,
        figsize=None,
        out_shape=(500, 500),
        title_fontsize=8,
        label_fontsize=6,
        legend_fontsize=6,
        names=None,
        fig_kwds=None,
        legend_kwds=None,
        subplots_kwds=None,
    ):
        """Plot a Raster object as a raster matrix

        Parameters
        ----------
        cmap : str (opt), default=None
            Specify a single cmap to apply to all of the RasterLayers.
            This overides the cmap attribute of each RasterLayer.

        norm :  matplotlib.colors.Normalize (opt), default=None
            A matplotlib.colors.Normalize to apply to all of the
            RasterLayers. This overides the norm attribute of each
            RasterLayer.

        figsize : tuple (opt), default=None
            Size of the resulting matplotlib.figure.Figure.

        out_shape : tuple, default=(500, 500)
            Number of rows, cols to read from the raster datasets for
            plotting.

        title_fontsize : any number, default=8
            Size in pts of titles.

        label_fontsize : any number, default=6
            Size in pts of axis ticklabels.

        legend_fontsize : any number, default=6
            Size in pts of legend ticklabels.

        names : list (opt), default=None
            Optionally supply a list of names for each RasterLayer to
            override the default layer names for the titles.

        fig_kwds : dict (opt), default=None
            Additional arguments to pass to the
            matplotlib.pyplot.figure call when creating the figure
            object.

        legend_kwds : dict (opt), default=None
            Additional arguments to pass to the
            matplotlib.pyplot.colorbar call when creating the colorbar
            object.

        subplots_kwds : dict (opt), default=None
            Additional arguments to pass to the
            matplotlib.pyplot.subplots_adjust function. These are used to
            control the spacing and position of each subplot, and can
            include{left=None, bottom=None, right=None, top=None,
            wspace=None, hspace=None}.

        Returns
        -------
        axs : numpy.ndarray
            array of matplotlib.axes._subplots.AxesSubplot or a single
            matplotlib.axes._subplots.AxesSubplot if Raster object
            contains only a single layer.
        """

        # some checks
        if norm:
            if not isinstance(norm, mpl.colors.Normalize):
                raise AttributeError(
                    "norm argument should be a matplotlib.colors.Normalize object"
                )

        if cmap:
            cmaps = [cmap for i in self.iloc]
        else:
            cmaps = [i.cmap for i in self.iloc]

        if norm:
            norms = [norm for i in self.iloc]
        else:
            norms = [i.norm for i in self.iloc]

        if names is None:
            names = self.names
        else:
            if len(names) != self.count:
                raise AttributeError(
                    "arguments 'names' needs to be the same length as the number of "
                    "RasterLayer objects "
                )

        if fig_kwds is None:
            fig_kwds = {}

        if legend_kwds is None:
            legend_kwds = {}

        if subplots_kwds is None:
            subplots_kwds = {}

        if figsize:
            fig_kwds["figsize"] = figsize

        # estimate required number of rows and columns in figure
        rows = int(np.sqrt(self.count))
        cols = int(math.ceil(np.sqrt(self.count)))

        if rows * cols < self.count:
            rows += 1

        fig, axs = plt.subplots(rows, cols, **fig_kwds)

        # axs.flat is an iterator over the row-order flattened axs array
        if isinstance(axs, np.ndarray):
            for ax, n, cmap, norm, name in zip(
                axs.flat, range(self.count), cmaps, norms, names
            ):

                arr = self.iloc[n].read(masked=True, out_shape=out_shape)
                ax.set_title(name, fontsize=title_fontsize, y=1.00)

                im = ax.imshow(
                    arr,
                    extent=[
                        self.bounds.left,
                        self.bounds.right,
                        self.bounds.bottom,
                        self.bounds.top,
                    ],
                    cmap=cmap,
                    norm=norm,
                )

                divider = make_axes_locatable(ax)

                if "orientation" not in legend_kwds.keys():
                    legend_kwds["orientation"] = "vertical"

                if legend_kwds["orientation"] == "vertical":
                    legend_pos = "right"

                elif legend_kwds["orientation"] == "horizontal":
                    legend_pos = "bottom"

                cax = divider.append_axes(legend_pos, size="10%", pad=0.1)
                cbar = plt.colorbar(im, cax=cax, **legend_kwds)
                cbar.ax.tick_params(labelsize=legend_fontsize)

                # hide tick labels by default when multiple rows or cols
                ax.axes.get_xaxis().set_ticklabels([])
                ax.axes.get_yaxis().set_ticklabels([])

                # show y-axis tick labels on first subplot
                if n == 0 and rows > 1:
                    ticks_loc = ax.get_yticks().tolist()
                    ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                    ax.set_yticklabels(
                        ax.yaxis.get_majorticklocs().astype("int"),
                        fontsize=label_fontsize,
                    )

                if n == 0 and rows == 1:
                    ticks_loc = ax.get_xticks().tolist()
                    ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                    ax.set_xticklabels(
                        ax.xaxis.get_majorticklocs().astype("int"),
                        fontsize=label_fontsize,
                    )

                    ticks_loc = ax.get_yticks().tolist()
                    ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                    ax.set_yticklabels(
                        ax.yaxis.get_majorticklocs().astype("int"),
                        fontsize=label_fontsize,
                    )

                if rows > 1 and n == (rows * cols) - cols:
                    ticks_loc = ax.get_xticks().tolist()
                    ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                    ax.set_xticklabels(
                        ax.xaxis.get_majorticklocs().astype("int"),
                        fontsize=label_fontsize,
                    )

            for ax in axs.flat[axs.size - 1 : self.count - 1 : -1]:
                ax.set_visible(False)

            plt.subplots_adjust(**subplots_kwds)

        else:
            arr = self.iloc[0].read(masked=True, out_shape=out_shape)
            cmap = cmaps[0]
            norm = norms[0]
            axs.set_title(list(names)[0], fontsize=title_fontsize, y=1.00)
            im = axs.imshow(
                arr,
                extent=[
                    self.bounds.left,
                    self.bounds.right,
                    self.bounds.bottom,
                    self.bounds.top,
                ],
                cmap=cmap,
                norm=norm,
            )

            divider = make_axes_locatable(axs)

            if "orientation" not in legend_kwds.keys():
                legend_kwds["orientation"] = "vertical"

            if legend_kwds["orientation"] == "vertical":
                legend_pos = "right"

            elif legend_kwds["orientation"] == "horizontal":
                legend_pos = "bottom"

            cax = divider.append_axes(legend_pos, size="10%", pad=0.1)
            cbar = plt.colorbar(im, cax=cax, **legend_kwds)
            cbar.ax.tick_params(labelsize=legend_fontsize)

        return axs



from functools import partial

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import rasterio
from rasterio.io import MemoryFile

from ._plotting import discrete_cmap
from ._rasterbase import get_nodata_value, _make_name


class RasterLayer:
    """Represents a single raster band derived from a single or
    multi-band raster dataset

    Simple wrapper around a rasterio.Band object with additional
    methods. Used because the Rasterio.Band.ds.read method reads
    all bands from a multi-band dataset, whereas the RasterLayer read
    method only reads a single band.

    Methods encapsulated in RasterLayer objects represent those that
    typically would only be applied to a single-band of a raster, i.e.
    sieve-clump, distance to non-NaN pixels, or arithmetic operations
    on individual layers.

    Attributes
    ----------
    bidx : int
        The band index of the RasterLayer within the file dataset.

    dtype : str
        The data type of the RasterLayer.

    ds : rasterio.band
        The underlying rasterio.band object.

    name : str
        A syntactically valid name for the RasterLayer.

    file : str
        The file path to the dataset.

    nodata : any number
        The number that is used to represent nodata pixels in the
        RasterLayer.

    driver : str
        The name of the GDAL format driver.

    meta : dict
        A python dict storing the RasterLayer metadata.

    transform : affine.Affine object
        The affine transform parameters.

    count : int
        Number of layers; always equal to 1.

    shape: tuple
        Shape of RasterLayer in (rows, columns)

    width, height: int
        The width (cols) and height (rows) of the dataset.

    bounds : BoundingBox named tuple
        A named tuple with left, bottom, right and top coordinates of
        the dataset.

    cmap : str
        The name of matplotlib map, or a custom
        matplotlib.cm.LinearSegmentedColormap or ListedColormap object.

    norm : matplotlib.colors.Normalize (opt)
        A matplotlib.colors.Normalize to apply to the RasterLayer.
        This overides the norm attribute of the RasterLayer.
    """

    def __init__(self, band):
        """Initiate a RasterLayer object

        Parameters
        ----------
        band : a rasterio.Band object
        """
        self.bidx = band.bidx
        self.dtype = band.dtype
        self.ds = band.ds

        if len(band.ds.files) > 0:
            description = band.ds.descriptions[band.bidx-1]
            if description is not None:
                layer_name = _make_name(band.ds.descriptions[band.bidx-1])
            else:
                layer_name = _make_name(band.ds.files[0])

            self.name = layer_name
            self.file = band.ds.files[0]

        else:
            self.name = "in_memory"
            self.file = None

        self.nodata = band.ds.nodata
        self.driver = band.ds.meta["driver"]
        self.meta = band.ds.meta
        self.transform = band.ds.transform
        self.crs = band.ds.crs
        self.count = 1
        self.shape = band.shape
        self.width = band.ds.width
        self.height = band.ds.height
        self.bounds = band.ds.bounds
        self.in_memory = False

        self.cmap = "viridis"
        self.norm = None
        self.categorical = False

    def close(self):
        self.ds.close()

    def _arith(self, function, other=None):
        """General method for performing arithmetic operations on
        RasterLayer objects

        Parameters
        ----------
        function : function
            Custom function that takes either one or two arrays, and
            returns a single array following a pre-defined calculation.

        other : pyspatialml.RasterLayer (optional, default None)
            If not specified, then a `function` should be provided that
            performs a calculation using only the selected RasterLayer.
            If `other` is specified, then a `function` should be
            supplied that takes to ndarrays as arguments and performs a
            calculation using both layers, i.e. layer1 - layer2.

        Returns
        -------
        pyspatialml.RasterLayer
            Returns a single RasterLayer containing the calculated
            result.
        """

        driver = self.driver

        if isinstance(other, RasterLayer):
            result = function(self.read(masked=True), other.read(masked=True))
        else:
            result = function(self.read(masked=True))

        nodata = get_nodata_value(result.dtype)

        # open output file with updated metadata
        meta = self.meta.copy()
        meta.update(driver=driver, count=1, dtype=result.dtype, nodata=nodata)

        with MemoryFile() as memfile:
            dst = memfile.open(**meta)
            result = np.ma.filled(result, fill_value=nodata)
            dst.write(result, indexes=1)

        # create RasterLayer from result
        layer = RasterLayer(rasterio.band(dst, 1))

        return layer

    def __add__(self, other):
        """Implements behaviour for addition of two RasterLayers,
        i.e. added_layer = layer1 + layer2
        """

        def func(arr1, arr2):
            return arr1 + arr2

        return self._arith(func, other)

    def __sub__(self, other):
        """Implements behaviour for subtraction of two RasterLayers, i.e.
        subtracted_layer = layer1 - layer2
        """

        def func(arr1, arr2):
            return arr1 - arr2

        return self._arith(func, other)

    def __mul__(self, other):
        """Implements behaviour for multiplication of two RasterLayers, i.e.
        product = layer1 * layer2
        """

        def func(arr1, arr2):
            return arr1 * arr2

        return self._arith(func, other)

    def __truediv__(self, other):
        """Implements behaviour for division using `/` of two RasterLayers,
        i.e. div = layer1 / layer2
        """

        def func(arr1, arr2):
            return arr1 / arr2

        return self._arith(func, other)

    def __and__(self, other):
        """Implements & operator

        Equivalent to a intersection operation of self
        with other, i.e. intersected = layer1 & layer2.
        """

        def func(arr1, arr2):
            mask = np.logical_and(arr1, arr2).mask
            arr1.mask[mask] = True
            return arr1

        return self._arith(func, other)

    def __or__(self, other):
        """Implements | operator

        Fills gaps in self with pixels from other. Equivalent to a union
        operation, i.e. union = layer1 | layer2.
        """

        def func(arr1, arr2):
            idx = np.logical_or(arr1, arr2.mask).mask
            arr1[idx] = arr2[idx]
            return arr1

        return self._arith(func, other)

    def __xor__(self, other):
        """Exclusive OR using ^

        Equivalent to a symmetrical difference where the result comprises
        pixels that occur in self or other, but not both, i.e.
        xor = layer1 ^ layer2.
        """

        def func(arr1, arr2):
            mask = ~np.logical_xor(arr1, arr2)
            idx = np.logical_or(arr1, arr2.mask).mask
            arr1[idx] = arr2[idx]
            arr1.mask[np.nonzero(mask)] = True
            return arr1

        return self._arith(func, other)

    def __round__(self, ndigits):
        """Behaviour for round() function, i.e. round(layer)"""

        def func(arr, ndigits):
            return np.round(arr, ndigits)

        func = partial(func, ndigits=ndigits)

        return self._arith(func)

    def __floor__(self):
        """Rounding down to the nearest integer using math.floor(),
        i.e. math.floor(layer)"""

        def func(arr):
            return np.floor(arr)

        return self._arith(func)

    def __ceil__(self):
        """Rounding up to the nearest integer using math.ceil(), i.e.
        math.ceil(layer)"""

        def func(arr):
            return np.ceil(arr)

        return self._arith(func)

    def __trunc__(self):
        """Truncating to an integral using math.trunc(), i.e.
        math.trunc(layer)"""

        def func(arr):
            return np.trunc(arr)

        return self._arith(func)

    def __abs__(self):
        """abs() function as applied to a RasterLayer, i.e. abs(layer)"""

        def func(arr):
            return np.abs(arr)

        return self._arith(func)

    def __pos__(self):
        """Unary positive, i.e. +layer1"""

        def func(arr):
            return np.positive(arr)

        return self._arith(func)

    def __neg__(self):
        """
        Unary negative, i.e. -layer1
        """

        def func(arr):
            return np.negative(arr)

        return self._arith(func)

    def _stats(self, max_pixels):
        """Take a sample of pixels from which to derive per-band
        statistics."""

        rel_width = self.shape[1] / max_pixels

        if rel_width > 1:
            col_scaling = round(max_pixels / rel_width)
            row_scaling = max_pixels - col_scaling
        else:
            col_scaling = round(max_pixels * rel_width)
            row_scaling = max_pixels - col_scaling

        out_shape = (row_scaling, col_scaling)
        arr = self.read(masked=True, out_shape=out_shape)
        arr = arr.flatten()
        return arr

    def min(self, max_pixels=10000):
        """Minimum value.

        Parameters
        ----------
        max_pixels : int
            Number of pixels used to inform statistical estimate.

        Returns
        -------
        numpy.float32
            The minimum value of the object
        """
        arr = self._stats(max_pixels)
        return np.nanmin(arr)

    def max(self, max_pixels=10000):
        """Maximum value.

        Parameters
        ----------
        max_pixels : int
            Number of pixels used to inform statistical estimate.

        Returns
        -------
        numpy.float32
            The maximum value of the object's pixels.
        """
        arr = self._stats(max_pixels)
        return np.nanmax(arr)

    def mean(self, max_pixels=10000):
        """Mean value

        Parameters
        ----------
        max_pixels : int
            Number of pixels used to inform statistical estimate.

        Returns
        -------
        numpy.float32
            The mean value of the object's pixels.
        """
        arr = self._stats(max_pixels)
        return np.nanmean(arr)

    def median(self, max_pixels=10000):
        """Median value

        Parameters
        ----------
        max_pixels : int
            Number of pixels used to inform statistical estimate.

        Returns
        -------
        numpy.float32
            The medium value of the object's pixels.
        """
        arr = self._stats(max_pixels)
        return np.nanmedian(arr)

    def stddev(self, max_pixels=10000):
        """Standard deviation

        Parameters
        ----------
        max_pixels : int
            Number of pixels used to inform statistical estimate.

        Returns
        -------
        numpy.float32
            The standard deviation of the object's pixels.
        """
        arr = self._stats(max_pixels)
        return np.nanstd(arr)

    def read(self, **kwargs):
        """Read method for a single RasterLayer.

        Reads the pixel values from a RasterLayer into a ndarray that
        always will have two dimensions in the order of (rows, columns).

        Parameters
        ----------
        **kwargs : named arguments that can be passed to the the
        rasterio.DatasetReader.read method.
        """
        if "resampling" in kwargs.keys():
            resampling_methods = [i.name for i in rasterio.enums.Resampling]

            if kwargs["resampling"] not in resampling_methods:
                raise ValueError(
                    "Invalid resampling method. Resampling "
                    "method must be one of {0}:".format(resampling_methods)
                )

            kwargs["resampling"] = rasterio.enums.Resampling[kwargs["resampling"]]

        return self.ds.read(indexes=self.bidx, **kwargs)

    def write(self, file_path, driver="GTiff", dtype=None, nodata=None, **kwargs):
        """Write method for a single RasterLayer.

        Parameters
        ----------
        file_path : str (opt)
            File path to save the dataset.

        driver : str
            GDAL-compatible driver used for the file format.

        dtype : str (opt)
            Numpy dtype used for the file. If omitted then the
            RasterLayer's dtype is used.

        nodata : any number (opt)
            A value used to represent the nodata pixels. If omitted
            then the RasterLayer's nodata value is used (if assigned
            already).

        kwargs : opt
            Optional named arguments to pass to the format drivers.
            For example can be `compress="deflate"` to add compression.

        Returns
        -------
        pyspatialml.RasterLayer
        """
        if dtype is None:
            dtype = self.dtype

        if nodata is None:
            nodata = get_nodata_value(dtype)

        meta = self.ds.meta
        meta["driver"] = driver
        meta["nodata"] = nodata
        meta["dtype"] = dtype
        meta.update(kwargs)

        # mask any nodata values
        arr = np.ma.masked_equal(self.read(), self.nodata)
        arr = arr.filled(fill_value=nodata)

        # write to file
        with rasterio.open(file_path, mode="w", **meta) as dst:
            dst.write(arr.astype(dtype), 1)

        src = rasterio.open(file_path)
        band = rasterio.band(src, 1)
        layer = RasterLayer(band)

        return layer

    def _extract_by_indices(self, rows, cols):
        """Spatial query of Raster object (by-band)"""

        X = np.ma.zeros((len(rows), self.count), dtype="float32")
        arr = self.read(masked=True)
        X[:, 0] = arr[rows, cols]

        return X

    def plot(
        self,
        cmap=None,
        norm=None,
        ax=None,
        cax=None,
        figsize=None,
        out_shape=(500, 500),
        categorical=None,
        legend=False,
        vmin=None,
        vmax=None,
        fig_kwds=None,
        legend_kwds=None,
    ):
        """Plot a RasterLayer using matplotlib.pyplot.imshow
        Parameters
        ----------
        cmap : str (default None)
            The name of a colormap recognized by matplotlib.
            Overrides the cmap attribute of the RasterLayer.

        norm : matplotlib.colors.Normalize (opt)
            A matplotlib.colors.Normalize to apply to the RasterLayer.
            This overrides the norm attribute of the RasterLayer.

        ax : matplotlib.pyplot.Artist (optional, default None)
            axes instance on which to draw to plot.

        cax : matplotlib.pyplot.Artist (optional, default None)
            axes on which to draw the legend.

        figsize : tuple of integers (optional, default None)
            Size of the matplotlib.figure.Figure. If the ax argument is
            given explicitly, figsize is ignored.

        out_shape : tuple, default=(500, 500)
            Number of rows, cols to read from the raster datasets for
            plotting.

        categorical : bool (optional, default False)
            if True then the raster values will be considered to
            represent discrete values, otherwise they are considered to
            represent continuous values. This overrides the
            RasterLayer 'categorical' attribute. Setting the argument
            categorical to True is ignored if the
            RasterLayer.categorical is already True.

        legend : bool (optional, default False)
            Whether to plot the legend.

        vmin, xmax : scale (optional, default None)
            vmin and vmax define the data range that the colormap
            covers. By default, the colormap covers the complete value
            range of the supplied data. vmin, vmax are ignored if the
            norm parameter is used.

        fig_kwds : dict (optional, default None)
            Additional arguments to pass to the
            matplotlib.pyplot.figure call when creating the figure
            object. Ignored if ax is passed to the plot function.

        legend_kwds : dict (optional, default None)
            Keyword arguments to pass to matplotlib.pyplot.colorbar().

        Returns
        -------
        ax : matplotlib axes instance

        """

        # some checks
        if fig_kwds is None:
            fig_kwds = {}

        if ax is None:
            if cax is not None:
                raise ValueError("'ax' can not be None if 'cax' is not.")
            fig, ax = plt.subplots(figsize=figsize, **fig_kwds)

        ax.set_aspect("equal")

        if norm:
            if not isinstance(norm, mpl.colors.Normalize):
                raise AttributeError(
                    "norm argument should be a " "matplotlib.colors.Normalize object"
                )

        if cmap is None:
            cmap = self.cmap

        if norm is None:
            norm = self.norm

        if legend_kwds is None:
            legend_kwds = {}

        arr = self.read(masked=True, out_shape=out_shape)

        if categorical is True:
            if self.categorical is False:
                N = np.bincount(arr)
                cmap = discrete_cmap(N, base_cmap=cmap)
            vmin, vmax = None, None

        im = ax.imshow(
            X=arr,
            extent=rasterio.plot.plotting_extent(self.ds),
            cmap=cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax,
        )

        if legend is True:
            plt.colorbar(im, cax=cax, ax=ax, **legend_kwds)

        return ax



from copy import deepcopy
from unittest import TestCase

import geopandas
import pandas as pd
import rasterio

from pyspatialml import Raster
from pyspatialml.datasets import nc


class TestExtract(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.extracted_grass = pd.read_table(nc.extracted_pixels, delimiter=" ")

        self.stack = Raster(self.predictors)

    def tearDown(self) -> None:
        self.stack.close()

    def test_extract_points(self):
        training_pt = geopandas.read_file(nc.points)

        # check that extracted training data as a DataFrame match known values
        df = self.stack.extract_vector(gdf=training_pt)
        df = df.dropna()
        training_pt = training_pt.dropna()

        self.assertTrue(
            (df["lsat7_2000_10"].values == training_pt["b1"].values).all()
        )
        self.assertTrue(
            (df["lsat7_2000_20"].values == training_pt["b2"].values).all()
        )
        self.assertTrue(
            (df["lsat7_2000_30"].values == training_pt["b3"].values).all()
        )
        self.assertTrue(
            (df["lsat7_2000_40"].values == training_pt["b4"].values).all()
        )
        self.assertTrue(
            (df["lsat7_2000_50"].values == training_pt["b5"].values).all()
        )
        self.assertTrue(
            (df["lsat7_2000_70"].values == training_pt["b7"].values).all()
        )

    def test_extract_polygons(self):
        # extract training data from polygons
        training_py = geopandas.read_file(nc.polygons)
        df = self.stack.extract_vector(gdf=training_py)
        df = df.dropna()

        df = df.merge(
            right=training_py.loc[:, ("id", "label")],
            left_on="geometry_idx",
            right_on="index",
            right_index=True,
        )

        # compare to extracted data using GRASS GIS
        self.assertEqual(df.shape[0], self.extracted_grass.shape[0])
        self.assertAlmostEqual(
            df["lsat7_2000_10"].mean(), self.extracted_grass["b1"].mean(),
            places=2
        )
        self.assertAlmostEqual(
            df["lsat7_2000_20"].mean(), self.extracted_grass["b2"].mean(),
            places=2
        )
        self.assertAlmostEqual(
            df["lsat7_2000_30"].mean(), self.extracted_grass["b3"].mean(),
            places=2
        )
        self.assertAlmostEqual(
            df["lsat7_2000_40"].mean(), self.extracted_grass["b4"].mean(),
            places=2
        )
        self.assertAlmostEqual(
            df["lsat7_2000_50"].mean(), self.extracted_grass["b5"].mean(),
            places=2
        )
        self.assertAlmostEqual(
            df["lsat7_2000_70"].mean(), self.extracted_grass["b7"].mean(),
            places=2
        )

    def test_extract_lines(self):
        # extract training data from lines
        training_py = geopandas.read_file(nc.polygons)
        training_lines = deepcopy(training_py)
        training_lines["geometry"] = training_lines.geometry.boundary
        df = self.stack.extract_vector(gdf=training_lines).dropna()

        # check shapes of extracted pixels
        self.assertEqual(df.shape[0], 948)

    def test_extract_raster(self):
        # extract training data from labelled pixels
        with rasterio.open(nc.labelled_pixels) as src:
            df = self.stack.extract_raster(src)

        df = df.dropna()

        self.assertEqual(df.shape[0], self.extracted_grass.shape[0])
        self.assertAlmostEqual(
            df["lsat7_2000_10"].mean(), self.extracted_grass["b1"].mean(),
            places=3
        )
        self.assertAlmostEqual(
            df["lsat7_2000_20"].mean(), self.extracted_grass["b2"].mean(),
            places=3
        )
        self.assertAlmostEqual(
            df["lsat7_2000_30"].mean(), self.extracted_grass["b3"].mean(),
            places=3
        )
        self.assertAlmostEqual(
            df["lsat7_2000_40"].mean(), self.extracted_grass["b4"].mean(),
            places=3
        )
        self.assertAlmostEqual(
            df["lsat7_2000_50"].mean(), self.extracted_grass["b5"].mean(),
            places=3
        )
        self.assertAlmostEqual(
            df["lsat7_2000_70"].mean(), self.extracted_grass["b7"].mean(),
            places=3
        )



import numpy as np
import unittest
import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestStats(unittest.TestCase):
    def setUp(self) -> None:
        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.predictors = predictors
        self.stack = Raster(predictors)

    def test_rasterstats(self):
        self.assertEqual(len(self.stack.min()), len(self.predictors))
        self.assertTrue(~np.isnan(self.stack.min()).all())

        self.assertEqual(len(self.stack.max()), len(self.predictors))
        self.assertTrue(~np.isnan(self.stack.max()).all())

        self.assertEqual(len(self.stack.mean()), len(self.predictors))
        self.assertTrue(~np.isnan(self.stack.mean()).all())

        self.assertEqual(len(self.stack.median()), len(self.predictors))
        self.assertTrue(~np.isnan(self.stack.median()).all())

    def test_layerstats(self):
        self.assertEqual(self.stack.iloc[0].min(), 56.0)
        self.assertEqual(self.stack.iloc[0].max(), 255.0)
        self.assertAlmostEqual(self.stack.iloc[0].mean(), 80.6, places=0)
        self.assertEqual(self.stack.iloc[0].median(), 75.0)

if __name__ == '__main__':
    unittest.main()



import multiprocessing
import os
import re
from collections import Counter

import numpy as np
import tempfile


def get_nodata_value(dtype):
    """Get a nodata value based on the minimum value permissible by dtype

    Parameters
    ----------
    dtype : str or dtype
        dtype to return a nodata value for

    Returns
    -------
    nodata : any number
        A nodata value that is accomodated by the supplied dtype
    """
    try:
        nodata = np.iinfo(dtype).min
    except ValueError:
        nodata = np.finfo(dtype).min

    return nodata


def get_num_workers(n_jobs):
    """Determine cpu count using scikit-learn convention of -1, -2 ...

    Parameters
    ----------
    n_jobs : int
        Number of processing cores including -1 for all cores -1, etc.

    Returns
    -------
    n_jobs : int
        The actual number of processing cores.
    """
    n_cpus = multiprocessing.cpu_count()

    if n_jobs < 0:
        n_jobs = n_cpus + n_jobs + 1

    return n_jobs


def _check_alignment(layers):
    """Check that a list of raster datasets are aligned with the same
    pixel dimensions and geotransforms.

    Parameters
    ----------
    layers : list
        List of pyspatialml.RasterLayer objects.

    Returns
    -------
    dict or False
        Dict of metadata if all layers are spatially aligned, otherwise
        returns False.
    """

    src_meta = []
    for layer in layers:
        src_meta.append(layer.ds.meta.copy())

    if not all(i["crs"] == src_meta[0]["crs"] for i in src_meta):
        Warning("crs of all rasters does not match, possible unintended consequences")

    if not all(
        [
            i["height"] == src_meta[0]["height"]
            or i["width"] == src_meta[0]["width"]
            or i["transform"] == src_meta[0]["transform"]
            for i in src_meta
        ]
    ):
        return False

    else:
        return src_meta[0]


def _make_name(name):
    """Converts a file basename to a valid class attribute name.

    Parameters
    ----------
    name : str
        File basename for converting to a valid class attribute name.

    Returns
    -------
    valid_name : str
        Syntactically correct name of layer so that it can form a class
        instance attribute.
    """
    basename = os.path.basename(name)
    sans_ext = os.path.splitext(basename)[0]

    valid_name = sans_ext.replace(" ", "_").replace("-", "_").replace(".", "_")

    if valid_name[0].isdigit():
        valid_name = "x" + valid_name

    valid_name = re.sub(r"[\[\]\(\)\{\}\;]", "", valid_name)
    valid_name = re.sub(r"_+", "_", valid_name)

    return valid_name


def _fix_names(combined_names):
    """Adjusts the names of pyspatialml.RasterLayer objects within the
    Raster when appending new layers.

    This avoids the Raster object containing duplicated names in the
    case that multiple RasterLayers are appended with the same name.

    In the case of duplicated names, the RasterLayer names are appended
    with a `_n` with n = 1, 2, 3 .. n.

    Parameters
    ----------
    combined_names : list
        List of str representing names of RasterLayers. Any duplicates
        will have a suffix appended to them.

    Returns
    -------
    list
        List with adjusted names
    """

    counts = Counter(combined_names)

    for s, num in counts.items():
        if num > 1:
            for suffix in range(1, num + 1):
                if s + "_" + str(suffix) not in combined_names:
                    combined_names[combined_names.index(s)] = s + "_" + str(suffix)
                else:
                    i = 1
                    while s + "_" + str(i) in combined_names:
                        i += 1
                    combined_names[combined_names.index(s)] = s + "_" + str(i)

    return combined_names


class TempRasterLayer:
    """Create a NamedTemporaryFile like object on Windows that has a
    close method

    Workaround used on Windows which cannot open the file a second time
    """

    def __init__(self, tempdir=tempfile.tempdir):
        self.tfile = tempfile.NamedTemporaryFile(dir=tempdir, suffix=".tif").name
        self.name = self.tfile

    def close(self):
        os.unlink(self.tfile)



import numpy as np


def extract_by_chunk(arr, w, idx, pixel_idx):
    d = idx.copy()
    pixel_idx = pixel_idx.copy()

    # subtract chunk offset from row, col positions
    d[:, 0] = d[:, 0] - w.row_off
    d[:, 1] = d[:, 1] - w.col_off

    # remove negative row, col positions
    pos = (d >= 0).all(axis=1)
    d = d[pos, :]
    pixel_idx = pixel_idx[pos]

    # remove row, col > shape
    within_range = (d[:, 0] < arr.shape[1]) & (d[:, 1] < arr.shape[2])
    d = d[within_range, :]
    pixel_idx = pixel_idx[within_range]

    extracted_data = arr[:, d[:, 0], d[:, 1]]

    return (extracted_data, pixel_idx)



from unittest import TestCase

import matplotlib as mpl
import numpy as np

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestPlotting(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.stack = Raster(self.predictors)
        self.stack_single = Raster(self.predictors[0])

    def tearDown(self) -> None:
        self.stack.close()
        self.stack_single.close()

    def test_plotting_raster(self):

        # test basic raster matrix plot
        p = self.stack.plot()
        self.assertIsInstance(p, np.ndarray)

        # test with arguments
        p = self.stack.plot(
            cmap="plasma",
            norm=mpl.colors.Normalize(vmin=10, vmax=100),
            title_fontsize=10,
            label_fontsize=10,
            names=["band1", "band2", "band3", "band4", "band5", "band7"],
            figsize=(10, 5),
            legend_kwds={"orientation": "horizontal"}
        )
        self.assertIsInstance(p, np.ndarray)

    def test_plotting_single(self):
        p = self.stack_single.plot(
            legend_kwds={"orientation": "horizontal", "fraction": 0.04})
        self.assertIsInstance(p, mpl.axes.Subplot)



from unittest import TestCase

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestDrop(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]

    def test_drop_inplace(self):
        stack = Raster(self.predictors)
        stack.drop(labels="lsat7_2000_50", in_place=True)

        # check that Raster object is returned
        self.assertIsInstance(stack, Raster)

        # check that RasterLayer has been dropped
        self.assertEqual(stack.count, 5)
        self.assertNotIn("lsat7_2000_50", stack.names)
        stack.close()

    def test_drop_with_copy(self):
        stack = Raster(self.predictors)
        names = stack.names
        result = stack.drop(labels="lsat7_2000_50", in_place=False)

        # check that Raster object is returned
        self.assertIsInstance(result, Raster)

        # check that RasterLayer has been dropped
        self.assertEqual(result.count, 5)
        self.assertNotIn("lsat7_2000_50", result.names)

        # check that original raster is unaffected
        self.assertEqual(stack.count, 6)
        self.assertEqual(stack.names, names)
        stack.close()
        result.close()

    def test_drop_in_memory(self):
        stack = Raster(self.predictors)
        names = stack.names

        result = stack.intersect(in_memory=True)
        result = stack.drop(labels="lsat7_2000_50", in_place=False)

        # check that Raster object is returned
        self.assertIsInstance(result, Raster)

        # check that RasterLayer has been dropped
        self.assertEqual(result.count, 5)
        self.assertNotIn("lsat7_2000_50", result.names)

        # check that original raster is unaffected
        self.assertEqual(stack.count, 6)
        self.assertEqual(stack.names, names)
        stack.close()



import os
from unittest import TestCase

import rasterio
import tempfile
import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestNames(TestCase):
    """Test the initiation of a Raster object when the file raster dataset
    contains band names
    """
    def setUp(self) -> None:
        """Create a temperory file with a raster dataset with band names
        """
        with rasterio.open(nc.multiband) as src:
            self.descriptions = ["band_1", "band_2", "band_3", "band_4", "band_5"]

            self.fp = tempfile.NamedTemporaryFile(suffix=".tif").name

            with rasterio.open(self.fp, "w", **src.meta) as dst:
                for i, band in enumerate(self.descriptions, start=1):
                    dst.write_band(i, src.read(i))
                    dst.set_band_description(i, band)

    def tearDown(self) -> None:
        os.remove(self.fp)

    def test_names_from_file(self) -> None:
        """Test the initiation of a Raster object from a file when the file raster 
        dataset contains band descriptions"""
        r = Raster(self.fp)
        self.assertEqual(list(r.names), self.descriptions)

    def test_names_from_rasterio(self) -> None:
        """Test the initiation of a Raster object from a rasterio.DatasetReader
        object when the file raster dataset has band descriptions
        """
        with rasterio.open(self.fp) as src:
            r = Raster(src)
            self.assertEqual(list(r.names), self.descriptions)

    def test_names_subsetting(self) -> None:
        """Test that the names of the bands are preserved when subsetting a raster
        """
        r = Raster(self.fp)
        subset = r.iloc[[0, 1]]
        self.assertEqual(list(subset.names), self.descriptions[0:2])

        new = r.copy(["band_1", "band_2"])
        new["band_3"] = r["band_3"]
        self.assertEqual(list(new.names), self.descriptions[0:3])



from unittest import TestCase

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestAppend(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]

    def test_append_inplace(self):
        """Append another Raster containing a single layer with identical name

        This test should cause the Raster object to automatically rename the
        duplicated names as "lsat7_2000_70_1", "lsat7_2000_70_2", etc.

        Appending a multi-band raster should result in a new layer with the
        multi-band name "landsat_multiband_1", "landsat_multiband_2", etc.

        A list of Rasters can be passed to append() to append multiple rasters
        """
        # append a single band raster with the same name
        stack = Raster(self.predictors)
        band7_mean = stack["lsat7_2000_70"].read(masked=True).mean()
        stack.append(Raster(nc.band7), in_place=True)

        self.assertEqual(list(stack.names)[5], "lsat7_2000_70_1")
        self.assertEqual(list(stack.names)[-1], "lsat7_2000_70_2")
        self.assertEqual(
            stack.lsat7_2000_70_1.read(masked=True).mean(),
            stack.lsat7_2000_70_2.read(masked=True).mean(),
            band7_mean,
        )

        # append a multiband raster
        stack = Raster(self.predictors)
        stack.append(Raster(nc.multiband), in_place=True)
        self.assertEqual(list(stack.names)[6], "landsat_multiband_1")
        stack.close()

        # append multiple rasters
        stack = Raster(self.predictors)
        stack.append([Raster(nc.band5), Raster(nc.band7)], in_place=True)
        self.assertEqual(stack.count, 8)

    def test_append_with_copy(self):
        """Same tests as above but create a new Raster rather than append
        in place
        """
        # append another Raster containing a single layer with identical name
        stack = Raster(self.predictors)
        band7_mean = stack["lsat7_2000_70"].read(masked=True).mean()
        result = stack.append(Raster(nc.band7), in_place=False)

        # check that original is untouched
        self.assertEqual(stack.count, 6)

        # check that result contains appended raster
        self.assertEqual(list(result.names)[5], "lsat7_2000_70_1")
        self.assertEqual(list(result.names)[-1], "lsat7_2000_70_2")

        # check that band 7 stats are the same after appending
        self.assertEqual(
            result.lsat7_2000_70_1.read(masked=True).mean(),
            result.lsat7_2000_70_2.read(masked=True).mean(),
            band7_mean,
        )

        # append a multiband raster
        result = stack.append(Raster(nc.multiband), in_place=False)
        self.assertEqual(list(result.names)[6], "landsat_multiband_1")
        stack.close()

        # append multiple rasters
        stack = Raster(self.predictors)
        new_stack = stack.append([Raster(nc.band5), Raster(nc.band7)], in_place=False)
        self.assertEqual(new_stack.count, 8)



from unittest import TestCase

import numpy as np

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestCalc(TestCase):
    def setUp(self) -> None:
        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
        self.stack = Raster(predictors)
        self.result = None

    def tearDown(self) -> None:
        self.stack.close()
        self.result.close()
        self.stack = None
        self.result = None

    def test_calc_with_2d_output(self):
        def compute_outputs_2d_array(arr):
            return arr[0, :, :] + arr[1, :, :]

        self.result = self.stack.apply(compute_outputs_2d_array)
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, 1)
        self.assertEqual(self.result.read(masked=True).count(), 183418)

    def test_calc_with_2d_output_coerce_dtype(self):
        def compute_outputs_2d_array(arr):
            return arr[0, :, :] + arr[1, :, :]

        self.result = self.stack.apply(compute_outputs_2d_array, dtype=np.int16)
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, 1)
        self.assertEqual(self.result.read(masked=True).count(), 183418)

    def test_calc_with_3d_output(self):
        def compute_outputs_3d_array(arr):
            arr[0, :, :] = arr[0, :, :] + arr[1, ::]
            return arr

        self.result = self.stack.apply(compute_outputs_3d_array)
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, 6)
        self.assertEqual(self.result.read(masked=True).count(), 1052182)

    def test_calc_with_multiprocessing(self):
        def compute_outputs_2d_array(arr):
            return arr[0, :, :] + arr[1, :, :]

        self.result = self.stack.apply(compute_outputs_2d_array)
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, 1)
        self.assertEqual(self.result.read(masked=True).count(), 183418)

    def test_calc_in_memory(self):
        def compute_outputs_2d_array(arr):
            return arr[0, :, :] + arr[1, :, :]

        self.result = self.stack.apply(compute_outputs_2d_array, in_memory=True)
        self.assertIsInstance(self.result, Raster)
        self.assertEqual(self.result.count, 1)
        self.assertEqual(self.result.read(masked=True).count(), 183418)



from unittest import TestCase

import os
import rasterio
import numpy as np

from pyspatialml import Raster, RasterLayer
from pyspatialml.datasets import nc


class TestInit(TestCase):
    def setUp(self) -> None:
        # inputs
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]

        # test results
        self.stack = None

    def tearDown(self) -> None:
        self.stack.close()

    def test_initiation_files(self):
        # test init from list of file paths
        self.stack = Raster(self.predictors)
        self.assertIsInstance(self.stack, Raster)
        self.assertEqual(self.stack.count, 6)

    def test_initiation_file(self):
        # test init from single file path
        self.stack = Raster(nc.band1)
        self.assertIsInstance(self.stack, Raster)
        self.assertEqual(self.stack.count, 1)

    def test_initiation_datasetreader(self):
        # test init from single rasterio.io.datasetreader
        with rasterio.open(nc.band1) as src:
            self.stack = Raster(src)
            self.assertIsInstance(self.stack, Raster)
            self.assertEqual(self.stack.count, 1)

    def test_initiation_list_datasetreader(self):
        # test init from list of rasterio.io.datasetreader objects
        srcs = []
        for f in self.predictors:
            srcs.append(rasterio.open(f))
        self.stack = Raster(srcs)
        self.assertIsInstance(self.stack, Raster)
        self.assertEqual(self.stack.count, 6)

    def test_initiation_band(self):
        # test init from single rasterio.band object
        with rasterio.open(nc.band1) as src:
            band = rasterio.band(src, 1)
            self.stack = Raster(band)
            self.assertIsInstance(self.stack, Raster)
            self.assertEqual(self.stack.count, 1)

    def test_initiation_list_bands(self):
        # test init from list of rasterio.band objects
        bands = []
        for f in self.predictors:
            src = rasterio.open(f)
            bands.append(rasterio.band(src, 1))
        self.stack = Raster(bands)
        self.assertIsInstance(self.stack, Raster)
        self.assertEqual(self.stack.count, 6)

    def test_initiation_rasterlayer(self):
        # test init from a single RasterLayer object
        with rasterio.open(nc.band1) as src:
            band = rasterio.band(src, 1)
            layer = RasterLayer(band)
            self.stack = Raster(layer)
            self.assertIsInstance(self.stack, Raster)
            self.assertEqual(self.stack.count, 1)

    def test_initiation_list_rasterlayer(self):
        # test init from a list of RasterLayer objects
        layers = []
        for f in self.predictors:
            src = rasterio.open(f)
            band = rasterio.band(src, 1)
            layers.append(RasterLayer(band))
        self.stack = Raster(layers)
        self.assertIsInstance(self.stack, Raster)
        self.assertEqual(self.stack.count, 6)

    def test_initiation_array(self):
        # check initiation of single-band raster from file
        arr = np.zeros((100, 100))
        self.stack = Raster(arr)
        
        # check output is written to tempfile
        self.assertTrue(os.path.exists(self.stack.iloc[0].file))

        # check some operations on the created raster
        layer_name = list(self.stack.names)[0]
        self.stack = self.stack.rename({layer_name: 'new_layer'})
        self.assertEqual(list(self.stack.names)[0], 'new_layer')

        # check initiation from array in memory
        arr = np.zeros((100, 100))
        self.stack = Raster(arr, in_memory=True)
        layer_name = list(self.stack.names)[0]

        self.stack = self.stack.rename({layer_name: 'new_layer'})
        self.assertEqual(list(self.stack.names)[0], 'new_layer')



import os

chnl_dist = os.path.join(os.path.dirname(__file__), 'chnl_dist.tif')
dem = os.path.join(os.path.dirname(__file__), 'dem.tif')
dist = os.path.join(os.path.dirname(__file__), 'dist.tif')
ffreq = os.path.join(os.path.dirname(__file__), 'ffreq.tif')
landimg2 = os.path.join(os.path.dirname(__file__), 'landimg2.tif')
landimg3 = os.path.join(os.path.dirname(__file__), 'landimg3.tif')
landimg4 = os.path.join(os.path.dirname(__file__), 'landimg4.tif')
mrvbf = os.path.join(os.path.dirname(__file__), 'mrvbf.tif')
rsp = os.path.join(os.path.dirname(__file__), 'rsp.tif')
slope = os.path.join(os.path.dirname(__file__), 'slope.tif')
soil = os.path.join(os.path.dirname(__file__), 'soil.tif')
twi = os.path.join(os.path.dirname(__file__), 'twi.tif')
meuse = os.path.join(os.path.dirname(__file__), 'meuse.shp')

predictors = [
    chnl_dist,
    dem,
    dist,
    ffreq,
    landimg2,
    landimg3,
    landimg4,
    mrvbf,
    rsp,
    slope,
    soil,
    twi]



from unittest import TestCase

import geopandas as gpd
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

import pyspatialml.datasets.meuse as ms
from pyspatialml import Raster
from pyspatialml.datasets import nc


class TestPrediction(TestCase):
    nc_predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]
    stack_nc = Raster(nc_predictors)
    stack_meuse = Raster(ms.predictors)

    def test_classification(self):
        training_pt = gpd.read_file(nc.points)
        df_points = self.stack_nc.extract_vector(gdf=training_pt)
        df_points["class_id"] = training_pt["id"].values
        df_points = df_points.dropna()

        clf = RandomForestClassifier(n_estimators=50)
        X = df_points.drop(columns=["class_id", "geometry"]).values
        y = df_points.class_id.values
        clf.fit(X, y)

        # classification
        cla = self.stack_nc.predict(estimator=clf, dtype="int16", nodata=0)
        self.assertIsInstance(cla, Raster)
        self.assertEqual(cla.count, 1)
        self.assertEqual(cla.read(masked=True).count(), 135092)

        # class probabilities
        probs = self.stack_nc.predict_proba(estimator=clf)
        self.assertIsInstance(cla, Raster)
        self.assertEqual(probs.count, 7)

        for layer in probs.values():
            self.assertEqual(layer.read(masked=True).count(), 135092)

    def test_classification_in_memory(self):
        training_pt = gpd.read_file(nc.points)
        df_points = self.stack_nc.extract_vector(gdf=training_pt)
        df_points["class_id"] = training_pt["id"].values
        df_points = df_points.dropna()

        clf = RandomForestClassifier(n_estimators=50)
        X = df_points.drop(columns=["class_id", "geometry"]).values
        y = df_points.class_id.values
        clf.fit(X, y)

        # classification
        cla = self.stack_nc.predict(estimator=clf, dtype="int16", nodata=0,
                                    in_memory=True)
        self.assertIsInstance(cla, Raster)
        self.assertEqual(cla.count, 1)
        self.assertEqual(cla.read(masked=True).count(), 135092)
        cla.close()

        # class probabilities
        probs = self.stack_nc.predict_proba(estimator=clf, in_memory=True)
        self.assertIsInstance(cla, Raster)
        self.assertEqual(probs.count, 7)

        for layer in probs.values():
            self.assertEqual(layer.read(masked=True).count(), 135092)

        probs.close()

    def test_regression(self):
        training_pt = gpd.read_file(ms.meuse)
        training = self.stack_meuse.extract_vector(gdf=training_pt)
        training["zinc"] = training_pt["zinc"].values
        training["cadmium"] = training_pt["cadmium"].values
        training["copper"] = training_pt["copper"].values
        training["lead"] = training_pt["lead"].values
        training = training.dropna()

        # single target regression
        regr = RandomForestRegressor(n_estimators=50)
        X = training.loc[:, self.stack_meuse.names].values
        y = training["zinc"].values
        regr.fit(X, y)

        single_regr = self.stack_meuse.predict(regr)
        self.assertIsInstance(single_regr, Raster)
        self.assertEqual(single_regr.count, 1)
        single_regr.close()

        # multi-target regression
        y = training.loc[:, ["zinc", "cadmium", "copper", "lead"]]
        regr.fit(X, y)
        multi_regr = self.stack_meuse.predict(regr)
        self.assertIsInstance(multi_regr, Raster)
        self.assertEqual(multi_regr.count, 4)
        multi_regr.close()

    def test_regression_in_memory(self):
        training_pt = gpd.read_file(ms.meuse)
        training = self.stack_meuse.extract_vector(gdf=training_pt)
        training["zinc"] = training_pt["zinc"].values
        training["cadmium"] = training_pt["cadmium"].values
        training["copper"] = training_pt["copper"].values
        training["lead"] = training_pt["lead"].values
        training = training.dropna()

        # single target regression
        regr = RandomForestRegressor(n_estimators=50)
        X = training.loc[:, self.stack_meuse.names].values
        y = training["zinc"].values
        regr.fit(X, y)

        single_regr = self.stack_meuse.predict(regr, in_memory=True)
        self.assertIsInstance(single_regr, Raster)
        self.assertEqual(single_regr.count, 1)
        single_regr.close()

        # multi-target regression
        y = training.loc[:, ["zinc", "cadmium", "copper", "lead"]]
        regr.fit(X, y)
        multi_regr = self.stack_meuse.predict(regr, in_memory=True)
        self.assertIsInstance(multi_regr, Raster)
        self.assertEqual(multi_regr.count, 4)
        multi_regr.close()

    def test_classification_with_single_constant(self):
        training_pt = gpd.read_file(nc.points)
        df_points = self.stack_nc.extract_vector(gdf=training_pt)
        df_points["class_id"] = training_pt["id"].values
        df_points = df_points.dropna()

        # classification with a single constant
        df_points["constant"] = 1

        clf = RandomForestClassifier(n_estimators=50)
        X = df_points.drop(columns=["class_id", "geometry"]).values
        y = df_points.class_id.values
        clf.fit(X, y)

        cla = self.stack_nc.predict(
            estimator=clf, 
            dtype="int16", 
            nodata=0,
            constants=[1]
        )
        self.assertIsInstance(cla, Raster)
        self.assertEqual(cla.count, 1)
        self.assertEqual(cla.read(masked=True).count(), 135092)

        probs = self.stack_nc.predict_proba(estimator=clf, constants=[1])
        self.assertIsInstance(cla, Raster)
        self.assertEqual(probs.count, 7)

        for layer in probs.values():
            self.assertEqual(layer.read(masked=True).count(), 135092)

    def test_classification_with_list_constants(self):
        training_pt = gpd.read_file(nc.points)
        df_points = self.stack_nc.extract_vector(gdf=training_pt)
        df_points["class_id"] = training_pt["id"].values
        df_points = df_points.dropna()

        df_points["constant"] = 1
        df_points["constant2"] = 2

        clf = RandomForestClassifier(n_estimators=50)
        X = df_points.drop(columns=["class_id", "geometry"]).values
        y = df_points.class_id.values
        clf.fit(X, y)

        cla = self.stack_nc.predict(
            estimator=clf, 
            dtype="int16", 
            nodata=0,
            constants=[1, 2]
        )
        self.assertIsInstance(cla, Raster)
        self.assertEqual(cla.count, 1)
        self.assertEqual(cla.read(masked=True).count(), 135092)

        probs = self.stack_nc.predict_proba(estimator=clf, constants=[1, 2])
        self.assertIsInstance(cla, Raster)
        self.assertEqual(probs.count, 7)

        for layer in probs.values():
            self.assertEqual(layer.read(masked=True).count(), 135092)

    def test_classification_with_dict_constants(self):
        # classification using constant to replace an existing layer
        training_pt = gpd.read_file(nc.points)
        df_points = self.stack_nc.extract_vector(gdf=training_pt)
        df_points["class_id"] = training_pt["id"].values
        df_points = df_points.dropna()

        clf = RandomForestClassifier(n_estimators=50)
        X = df_points.drop(columns=["class_id", "geometry"]).values
        y = df_points.class_id.values
        clf.fit(X, y)

        cla = self.stack_nc.predict(
            estimator=clf, 
            dtype="int16", 
            nodata=0,
            constants={"lsat7_2000_10": 150}
        )
        self.assertIsInstance(cla, Raster)
        self.assertEqual(cla.count, 1)
        self.assertEqual(cla.read(masked=True).count(), 135092)

        probs = self.stack_nc.predict_proba(estimator=clf, constants={"lsat7_2000_10": 150})
        self.assertIsInstance(cla, Raster)
        self.assertEqual(probs.count, 7)

        for layer in probs.values():
            self.assertEqual(layer.read(masked=True).count(), 135092)



from copy import deepcopy

import numpy as np
import rasterio
from scipy import ndimage

from .raster import Raster


def one_hot_encode(layer, file_path, categories=None, driver="GTiff"):
    """One-hot encoding of a RasterLayer.

    Parameters
    ----------
    layer : pyspatialml.RasterLayer
        Containing categories to perform one-hot encoding on.

    file_path : str
        File path to save one-hot encoded raster.

    categories : list, ndarray, optional
        Optional list of categories to extract. Default performs one-hot
        encoding on all categorical values in the input layer.

    driver : str, options. Default is 'GTiff'
        GDAL-compatible driver.

    Returns
    -------
    pyspatialml.Raster
        Each categorical value is encoded as a layer with a Raster object.
    """
    arr = layer.read(masked=True)

    if categories is None:
        categories = np.unique(arr)
        categories = categories[~categories.mask]
        categories = categories.data.astype("int32")

    arr_ohe = np.ma.zeros((len(categories), arr.shape[0], arr.shape[1]), dtype="int32")
    names = []
    prefix = layer.names[0]

    for i, cat in enumerate(categories):
        enc = deepcopy(arr)
        enc[enc != cat] = 0
        enc[enc == cat] = 1
        arr_ohe[i, :, :] = enc

        names.append("_".join([prefix, "cat", str(cat)]))

    # create new stack
    meta = deepcopy(layer.ds.meta)
    meta["driver"] = driver
    meta["nodata"] = -99999
    meta["count"] = arr_ohe.shape[0]
    meta["dtype"] = "int32"

    with rasterio.open(file_path, mode="w", **meta) as dst:
        dst.write(arr_ohe)

    new_raster = Raster(file_path)
    new_raster.rename({old: new for old, new in zip(new_raster.names, names)})

    return new_raster


def xy_coordinates(layer, file_path, driver="GTiff"):
    """
    Fill 2d arrays with their x,y indices.

    Parameters
    ----------
    layer : pyspatialml.RasterLayer, or rasterio.DatasetReader
        RasterLayer to use as a template.

    file_path : str
        File path to save to the resulting Raster object.s

    driver : str, options. Default is 'GTiff'
        GDAL driver to use to save raster.

    Returns
    -------
    pyspatialml.Raster object
    """

    arr = np.zeros(layer.shape, dtype=np.float32)
    arr = arr[np.newaxis, :, :]
    xyarrays = np.repeat(arr[0:1, :, :], 2, axis=0)
    xx, xy = np.meshgrid(np.arange(arr.shape[2]), np.arange(arr.shape[1]))
    xyarrays[0, :, :] = xx
    xyarrays[1, :, :] = xy

    # create new stack
    meta = deepcopy(layer.meta)
    meta["driver"] = driver
    meta["count"] = 2
    meta["dtype"] = xyarrays.dtype

    with rasterio.open(file_path, "w", **meta) as dst:
        dst.write(xyarrays)

    new_raster = Raster(file_path)
    names = ["x_coordinates", "y_coordinates"]
    new_raster.rename({old: new for old, new in zip(new_raster.names, names)})

    return new_raster


def rotated_coordinates(layer, file_path, n_angles=8, driver="GTiff"):
    """Generate 2d arrays with n_angles rotated coordinates.

    Parameters
    ----------
    layer : pyspatialml.RasterLayer, or rasterio.DatasetReader
        RasterLayer to use as a template.

    n_angles : int, optional. Default is 8
        Number of angles to rotate coordinate system by.

    driver : str, optional. Default is 'GTiff'
        GDAL driver to use to save raster.

    Returns
    -------
    pyspatialml.Raster
    """
    # define x and y grid dimensions
    xmin, ymin, xmax, ymax = 0, 0, layer.shape[1], layer.shape[0]
    x_range = np.arange(start=xmin, stop=xmax, step=1)
    y_range = np.arange(start=ymin, stop=ymax, step=1, dtype=np.float32)

    X_var, Y_var, _ = np.meshgrid(x_range, y_range, n_angles)
    angles = np.deg2rad(np.linspace(0, 180, n_angles, endpoint=False))
    grids_directional = X_var + np.tan(angles) * Y_var

    # reorder to band, row, col order
    grids_directional = grids_directional.transpose((2, 0, 1))

    # create new stack
    meta = deepcopy(layer.meta)
    meta["driver"] = driver
    meta["count"] = n_angles
    meta["dtype"] = grids_directional.dtype
    with rasterio.open(file_path, "w", **meta) as dst:
        dst.write(grids_directional)

    new_raster = Raster(file_path)
    names = ["angle_" + str(i + 1) for i in range(n_angles)]
    new_raster.rename({old: new for old, new in zip(new_raster.names, names)})

    return new_raster


def distance_to_corners(layer, file_path, driver="GTiff"):
    """Generate buffer distances to corner and centre coordinates of raster
    extent.

    Parameters
    ----------
    layer : pyspatialml.RasterLayer, or rasterio.DatasetReader

    file_path : str
        File path to save to the resulting Raster object

    driver : str, optional. Default is 'GTiff'
        GDAL driver to use to save raster.

    Returns
    -------
    pyspatialml.Raster object
    """

    names = ["top_left", "top_right", "bottom_left", "bottom_right", "centre_indices"]

    rows = np.asarray(
        [0, 0, layer.shape[0] - 1, layer.shape[0] - 1, int(layer.shape[0] / 2)]
    )
    cols = np.asarray(
        [0, layer.shape[1] - 1, 0, layer.shape[1] - 1, int(layer.shape[1] / 2)]
    )

    # euclidean distances
    arr = _grid_distance(layer.shape, rows, cols)

    # create new stack
    meta = deepcopy(layer.meta)
    meta["driver"] = driver
    meta["count"] = 5
    meta["dtype"] = arr.dtype

    with rasterio.open(file_path, "w", **meta) as dst:
        dst.write(arr)

    new_raster = Raster(file_path)
    new_raster.rename({old: new for old, new in zip(new_raster.names, names)})

    return new_raster


def _grid_distance(shape, rows, cols):
    """Generate buffer distances to x,y coordinates.
    Parameters
    ----------
    shape : tuple
        shape of numpy array (rows, cols) to create buffer distances within.
    rows : 1d numpy array
        array of row indexes.
    cols : 1d numpy array
        array of column indexes.
    Returns
    -------
    ndarray
        3d numpy array of euclidean grid distances to each x,y coordinate pair
        [band, row, col].
    """

    # create buffer distances
    grids_buffers = np.zeros((shape[0], shape[1], rows.shape[0]), dtype=np.float32)

    for i, (y, x) in enumerate(zip(rows, cols)):
        # create 2d array (image) with pick indexes set to z
        point_arr = np.zeros((shape[0], shape[1]))
        point_arr[y, x] = 1
        buffer = ndimage.morphology.distance_transform_edt(1 - point_arr)
        grids_buffers[:, :, i] = buffer

    # reorder to band, row, column
    grids_buffers = grids_buffers.transpose((2, 0, 1))

    return grids_buffers


def distance_to_samples(layer, file_path, rows, cols, driver="GTiff"):
    """Generate buffer distances to x,y coordinates.

    Parameters
    ----------
    layer : pyspatialml.RasterLayer, or rasterio.DatasetReader
        RasterLayer to use as a template.

    file_path : str
        File path to save to the resulting Raster object.

    rows : 1d numpy array
        array of row indexes.

    cols : 1d numpy array
        array of column indexes.

    driver : str, default='GTiff'
        GDAL driver to use to save raster.

    Returns
    -------
    pyspatialml.Raster object
    """
    # some checks
    if isinstance(rows, list):
        rows = np.asarray(rows)

    if isinstance(cols, list):
        cols = np.asarray(cols)

    if rows.shape != cols.shape:
        raise ValueError("rows and cols must have same dimensions")

    shape = layer.shape
    arr = _grid_distance(shape, rows, cols)

    # create new stack
    meta = deepcopy(layer.meta)
    meta["driver"] = driver
    meta["count"] = arr.shape[0]
    meta["dtype"] = arr.dtype

    with rasterio.open(file_path, "w", **meta) as dst:
        dst.write(arr)

    names = ["dist_sample" + str(i + 1) for i in range(len(rows))]
    new_raster = Raster(file_path)
    new_raster.rename({old: new for old, new in zip(new_raster.names, names)})

    return new_raster



from unittest import TestCase

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestToCrs(TestCase):
    def setUp(self) -> None:
        # test inputs
        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                      nc.band7]
        self.stack = Raster(predictors)

        # test results
        self.stack_prj = None

    def tearDown(self) -> None:
        self.stack.close()
        self.stack_prj.close()

    def test_to_crs_defaults(self):
        self.stack_prj = self.stack.to_crs({"init": "EPSG:4326"})

        # check raster object
        self.assertIsInstance(self.stack_prj, Raster)
        self.assertEqual(self.stack_prj.count, self.stack.count)
        self.assertEqual(self.stack_prj.read(masked=True).count(), 1012061)

        # test nodata value is recognized
        self.assertEqual(
            self.stack_prj.read(masked=True).min(),
            self.stack.read(masked=True).min()
        )
        self.assertEqual(
            self.stack_prj.read(masked=True).max(),
            self.stack.read(masked=True).max()
        )

    def test_to_crs_custom_nodata(self):
        self.stack_prj = self.stack.to_crs({"init": "EPSG:4326"}, nodata=-999)

        # check raster object
        self.assertIsInstance(self.stack_prj, Raster)
        self.assertEqual(self.stack_prj.count, self.stack.count)
        self.assertEqual(self.stack_prj.read(masked=True).count(), 1012061)

        # test nodata value is recognized
        self.assertEqual(
            self.stack_prj.read(masked=True).min(),
            self.stack.read(masked=True).min()
        )
        self.assertEqual(
            self.stack_prj.read(masked=True).max(),
            self.stack.read(masked=True).max()
        )

    def test_to_crs_in_memory(self):
        self.stack_prj = self.stack.to_crs({"init": "EPSG:4326"},
                                           in_memory=True)

        # check raster object
        self.assertIsInstance(self.stack_prj, Raster)



from unittest import TestCase

import geopandas as gpd
import numpy as np

import pyspatialml.datasets.nc as nc
from pyspatialml import Raster


class TestMask(TestCase):
    def setUp(self) -> None:
        # test inputs
        training_py = gpd.read_file(nc.polygons)
        self.mask_py = training_py.iloc[0:1, :]

        predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                      nc.band7]
        self.stack = Raster(predictors)

        # test results
        self.masked_object = None

    def tearDown(self) -> None:
        self.stack.close()
        self.masked_object.close()

    def test_mask_defaults(self):
        self.masked_object = self.stack.mask(self.mask_py)

        # check raster object
        self.assertIsInstance(self.masked_object, Raster)
        self.assertEqual(self.masked_object.count, self.stack.count)
        self.assertEqual(self.masked_object.read(masked=True).count(), 738)

        # test nodata value is recognized
        self.assertEqual(self.masked_object.read(masked=True).min(), 38.0)
        self.assertEqual(self.masked_object.read(masked=True).max(), 168.0)

    def test_mask_inverted(self):
        self.masked_object = self.stack.mask(self.mask_py, invert=True)

        # check raster object
        self.assertIsInstance(self.masked_object, Raster)
        self.assertEqual(self.masked_object.count, self.stack.count)
        self.assertEqual(self.masked_object.read(masked=True).count(), 1051444)

        # test nodata value is recognized
        self.assertEqual(self.masked_object.read(masked=True).min(), 1.0)
        self.assertEqual(self.masked_object.read(masked=True).max(), 255.0)

    def test_mask_custom_dtype(self):
        self.masked_object = self.stack.mask(self.mask_py, dtype=np.int16)

        # check raster object
        self.assertIsInstance(self.masked_object, Raster)
        self.assertEqual(self.masked_object.count, self.stack.count)
        self.assertEqual(self.masked_object.read(masked=True).count(), 738)

        # test nodata value is recognized
        self.assertEqual(self.masked_object.read(masked=True).min(), 38)
        self.assertEqual(self.masked_object.read(masked=True).max(), 168)

    def test_mask_custom_nodata(self):
        self.masked_object = self.stack.mask(self.mask_py, nodata=-99999)

        # check raster object
        self.assertIsInstance(self.masked_object, Raster)
        self.assertEqual(self.masked_object.count, self.stack.count)
        self.assertEqual(self.masked_object.read(masked=True).count(), 738)

        # test nodata value is recognized
        self.assertEqual(self.masked_object.read(masked=True).min(), 38.0)
        self.assertEqual(self.masked_object.read(masked=True).max(), 168.0)

    def test_mask_in_memory(self):
        self.masked_object = self.stack.mask(self.mask_py, in_memory=True)

        # check raster object
        self.assertIsInstance(self.masked_object, Raster)



from unittest import TestCase

from pyspatialml import Raster, RasterLayer
from pyspatialml.datasets import nc


class TestIndexing(TestCase):
    def setUp(self) -> None:
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5, nc.band7]

    def test_naming(self):
        stack = Raster(self.predictors + [nc.multiband])

        # check unique naming when stacking multiband raster
        self.assertEqual(stack.count, 11)
        expected_names = [
            "lsat7_2000_10",
            "lsat7_2000_20",
            "lsat7_2000_30",
            "lsat7_2000_40",
            "lsat7_2000_50",
            "lsat7_2000_70",
            "landsat_multiband_1",
            "landsat_multiband_2",
            "landsat_multiband_3",
            "landsat_multiband_4",
            "landsat_multiband_5",
        ]
        self.assertListEqual(list(stack.names), expected_names)
        stack.close()

    def test_subset_single_layer(self):
        stack = Raster(self.predictors + [nc.multiband])

        # Subset a single layer using an index position - returns a RasterLayer
        self.assertIsInstance(stack.iloc[0], RasterLayer)

        # Subset a single layer using a label - returns a RasterLayer
        self.assertIsInstance(stack["lsat7_2000_10"], RasterLayer)

        # Subset a single layer using an attribute - returns a RasterLayer
        self.assertIsInstance(stack.lsat7_2000_10, RasterLayer)

        # Check that the raster values are the same as the original values
        # after subsetting
        self.assertEqual(
            stack.lsat7_2000_10.read(masked=True).mean(),
            80.56715262406088
        )
        self.assertEqual(
            stack.lsat7_2000_70.read(masked=True).mean(),
            59.17773813401238
        )
        stack.close()

    def test_subset_multiple_layers(self):
        stack = Raster(self.predictors + [nc.multiband])

        # Subset multiple layers using a slice of index positions
        # - returns a Raster object
        self.assertIsInstance(stack.iloc[0:2], Raster)

        # Subset multiple layers using a list of index positions
        # - returns a Raster object
        self.assertIsInstance(stack.iloc[[0, 1, 2]], Raster)

        # Subset multiple layers using a list of labels
        # - returns a Raster object
        subset_raster = stack[["lsat7_2000_10", "lsat7_2000_70"]]
        self.assertIsInstance(subset_raster, Raster)
        self.assertListEqual(
            list(subset_raster.names),
            ["lsat7_2000_10", "lsat7_2000_70"]
        )

        # Check that label and integer subset return the same layers
        self.assertListEqual(
            list(stack.iloc[0:3].names),
            list(stack[["lsat7_2000_10", "lsat7_2000_20", "lsat7_2000_30"]].names),
        )

        stack.close()

    def test_indexing(self):
        stack = Raster(self.predictors + [nc.multiband])

        # replace band 1 with band 7
        band7_mean = stack["lsat7_2000_70"].read(masked=True).mean()

        stack.iloc[0] = Raster(nc.band7).iloc[0]

        self.assertEqual(stack.iloc[0].read(masked=True).mean(), band7_mean)
        self.assertEqual(stack["lsat7_2000_10"].read(masked=True).mean(),
                         band7_mean)
        self.assertEqual(stack["lsat7_2000_10"].read(masked=True).mean(),
                         band7_mean)
        self.assertEqual(stack.lsat7_2000_10.read(masked=True).mean(),
                         band7_mean)

        stack.close()



import random

from scipy.cluster.hierarchy import cut_tree, linkage
from shapely.geometry import Point


def filter_points(gdf, min_dist=0, remove="first"):
    """Filter points in geodataframe using a minimum distance buffer.

    Args
    ----
    gdf : Geopandas GeoDataFrame
        Containing point geometries.

    min_dist : int or float, optional (default=0)
        Minimum distance by which to filter out closely spaced points.

    remove : str, optional (default='first')
        Optionally choose to remove 'first' occurrences or 'last'
        occurrences.

    Returns
    -------
    xy : 2d array-like
        Numpy array filtered coordinates
    """
    xy = gdf.geometry.bounds.iloc[:, 0:2]

    Z = linkage(xy, "complete")
    tree_thres = cut_tree(Z, height=min_dist)
    gdf["tree_thres"] = tree_thres

    if remove == "first":
        gdf = gdf.groupby(by="tree_thres").first()

    elif remove == "last":
        gdf = gdf.groupby(by="tree_thres").last()

    return gdf


def get_random_point_in_polygon(poly):
    """Generates random shapely Point geometry objects within a single
    shapely Polygon object.

    Args
    ----
    poly : Shapely Polygon object

    Returns
    -------
    p : Shapely Point object
    """

    (minx, miny, maxx, maxy) = poly.bounds

    while True:
        p = Point(random.uniform(minx, maxx), random.uniform(miny, maxy))

        if poly.contains(p):
            return p



import numpy as np
from scipy.spatial.distance import cdist
from scipy.spatial import cKDTree
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import Normalizer
from sklearn.utils.extmath import weighted_mode


class KNNTransformer(BaseEstimator, TransformerMixin):
    """Transformer to generate new lag features by weighted aggregation
    of K-neighboring observations.

    A lag transformer uses a weighted mean/mode of the values of the
    K-neighboring observations to generate new lagged features. The
    weighted mean/mode of the surrounding observations are appended
    as a new feature to the right-most column in the training data.

    The K-neighboring observations are determined using the distance
    metric specified in the `metric` argument. The default metric is
    minkowski, and with p=2 is equivalent to the standard Euclidean
    metric.

    Parameters
    ----------
    n_neighbors : int, default = 7
        Number of neighbors to use by default for kneighbors queries.

    weights : {‘uniform’, ‘distance’} or callable, default=’distance’
        Weight function used in prediction. Possible values:

            - ‘uniform’ : uniform weights. All points in each
              neighborhood are weighted equally.
            - ‘distance’ : weight points by the inverse of their
               distance. In this case, closer neighbors of a query
               point will have a greater influence than neighbors
               which are further away.
            - [callable] : a user-defined function which accepts an
              array of distances, and returns an array of the same
              shape containing the weights.

    measure : {'mean', 'mode'}
        Function that is used to apply the weights to `y`. Use 'mean'
        if the target variable is continuous and 'mode' if the target
        variable is discrete.

    radius : float, default=1.0
        Range of parameter space to use by default for radius_neighbors
        queries.

    algorithm: {‘auto’, ‘ball_tree’, ‘kd_tree’, ‘brute’}, default=’auto’
        Algorithm used to compute the nearest neighbors:

            - ‘ball_tree’ will use BallTree
            - ‘kd_tree’ will use KDTree
            - ‘brute’ will use a brute-force search.
            - ‘auto’ will attempt to decide the most appropriate
               algorithm based on the values passed to fit method.
            - Note: fitting on sparse input will override the setting
              of this parameter, using brute force.

    leaf_size : int, default=30
        Leaf size passed to BallTree or KDTree. This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree. The optimal value depends on the
        nature of the problem.

    metric : str or callable, default=’minkowski’
        The distance metric to use for the tree. The default metric is
        minkowski, and with p=2 is equivalent to the standard
        Euclidean metric. See the documentation of DistanceMetric for
        a list of available metrics. If metric is “precomputed”, X is
        assumed to be a distance matrix and must be square during fit.
        X may be a sparse graph, in which case only “nonzero” elements
        may be considered neighbors.

    p : int, default=2
        Parameter for the Minkowski metric from
        sklearn.metrics.pairwise.pairwise_distances. When p = 1, this
        is equivalent to using manhattan_distance (l1), and
        euclidean_distance (l2) for p = 2. For arbitrary p,
        minkowski_distance (l_p) is used.

    normalize : bool, default=True
        Whether to normalize the inputs using
        sklearn.preprocessing.Normalizer

    metric_params : dict, default=None
        Additional keyword arguments for the metric function.

    kernel_params : dict, default=None
        Additional keyword arguments to pass to a custom kernel
        function.

    n_jobs : int, default=None
        The number of parallel jobs to run for neighbors search. None
        means 1 unless in a joblib.parallel_backend context. -1 means
        using all processors. See Glossary for more details.
    """

    def __init__(
        self,
        n_neighbors=7,
        weights="distance",
        measure="mean",
        radius=1.0,
        algorithm="auto",
        leaf_size=30,
        metric="minkowski",
        p=2,
        normalize=True,
        metric_params=None,
        kernel_params=None,
        n_jobs=1,
    ):

        self.n_neighbors = n_neighbors
        self.weights = weights
        self.measure = measure
        self.radius = radius
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.metric = metric
        self.p = p
        self.metric_params = metric_params
        self.kernel_params = kernel_params
        self.normalize = normalize
        self.n_jobs = n_jobs

        self.knn = NearestNeighbors(
            n_neighbors=self.n_neighbors,
            radius=self.radius,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric=self.metric,
            p=self.p,
            metric_params=self.metric_params,
            n_jobs=self.n_jobs,
        )

        self.y_ = None

    def fit(self, X, y=None):
        """Fit the base_estimator with features from X
        {n_samples, n_features} and with an additional spatially lagged
        variable added to the right-most column of the training data.

        During fitting, the k-neighbors to each training point are
        used to estimate the spatial lag component. The training point
        is not included in the calculation, i.e. the training point is
        not considered its own neighbor.

        Parameters
        ----------
        X : array-like of sample {n_samples, n_features} using for model
            fitting The training input samples

        y : array-like of shape (n_samples,)
            The target values (class labels in classification, real
            numbers in regression).
        """
        # some checks
        if self.kernel_params is None:
            self.kernel_params = {}

        if y.ndim == 1:
            self.n_outputs_ = 1
        else:
            self.n_outputs_ = y.shape[1]

        # fit knn and get values of neighbors
        if self.normalize is True:
            scaler = Normalizer()
            X = scaler.fit_transform(X)
            self.scaler_ = scaler

        self.knn.fit(X)
        self.y_ = y.copy()

        return self

    def transform(self, X, y=None):
        """Transform method for spatial lag models.

        Augments new observations with a spatial lag variable created
        from a weighted mean/mode (regression/classification) of
        k-neighboring observations.

        Parameters
        ----------
        X : array-like of sample {n_samples, n_features}
            New samples for the prediction.

        y : None
            Not used.
        """
        # get distances from training points to new data
        if self.normalize is True:
            X = self.scaler_.transform(X)

        neighbor_dist, neighbor_ids = self.knn.kneighbors(X=X)

        # mask zero distances
        neighbor_dist = np.ma.masked_equal(neighbor_dist, 0)

        # get values of closest training points to new data
        neighbor_vals = np.array([self.y_[i] for i in neighbor_ids])

        # mask neighbor values with zero distances
        mask = neighbor_dist.mask

        if mask.all() == False:
            mask = np.zeros(neighbor_dist.shape, dtype=np.bool)
            mask[:] = False

        if neighbor_vals.ndim == 2:
            neighbor_vals = np.ma.masked_array(neighbor_vals, mask)
        else:
            n_outputs = neighbor_vals.shape[2]
            mask = np.repeat(mask[:, :, np.newaxis], n_outputs, axis=2)
            neighbor_vals = np.ma.masked_array(neighbor_vals, mask=mask)

        # calculated weighted means
        if self.weights == "distance":
            new_X = self._distance_weighting(neighbor_vals, neighbor_dist)

        elif self.weights == "uniform":
            new_X = self._uniform_weighting(neighbor_vals)

        elif callable(self.weights):
            new_X = self._custom_weighting(neighbor_vals, neighbor_dist)

        return np.column_stack((X, new_X))

    def _apply_weights(self, neighbor_vals, neighbor_weights):
        # weighted mean/mode of neighbors for a single regression target
        if neighbor_vals.ndim == 2:
            if self.measure == "mean":
                X = np.ma.average(neighbor_vals, weights=neighbor_weights, axis=1)
            else:
                X, _ = weighted_mode(neighbor_vals, neighbor_weights, axis=1)

        # weighted mean of neighbors for a multi-target regression
        # neighbor_vals = (n_samples, n_neighbors, n_targets)
        else:
            X = np.zeros((neighbor_vals.shape[0], neighbor_vals.shape[2]))

            if self.measure == "mean":
                for i in range(neighbor_vals.shape[-1]):
                    X[:, i] = np.ma.average(
                        neighbor_vals[:, :, i], weights=neighbor_weights, axis=1
                    )
            else:
                for i in range(neighbor_vals.shape[-1]):
                    X[:, i], _ = weighted_mode(
                        neighbor_vals[:, :, i], neighbor_weights, axis=1
                    )

        return X

    def _distance_weighting(self, neighbor_vals, neighbor_dist):
        weights = 1 / neighbor_dist
        return self._apply_weights(neighbor_vals, weights)

    def _uniform_weighting(self, neighbor_vals):
        weights = np.ones((neighbor_vals.shape[0], neighbor_vals.shape[0]))
        return self._apply_weights(neighbor_vals, weights)

    def _custom_weighting(self, neighbor_vals, neighbor_dist):
        weights = self.weights(neighbor_dist, **self.kernel_params)
        return self._apply_weights(neighbor_vals, weights)

    def _distance_weighting(self, neighbor_vals, neighbor_dist):
        weights = 1 / neighbor_dist
        return self._apply_weights(neighbor_vals, weights)

    def _uniform_weighting(self, neighbor_vals):
        weights = np.ones((neighbor_vals.shape[0], neighbor_vals.shape[0]))
        return self._apply_weights(neighbor_vals, weights)

    def _custom_weighting(self, neighbor_vals, neighbor_dist):
        weights = self.weights(neighbor_dist, **self.kernel_params)
        return self._apply_weights(neighbor_vals, weights)


class GeoDistTransformer(BaseEstimator, TransformerMixin):
    """Transformer to add new features based on geographical distances
    to reference locations.

    Parameters
    ----------
    refs : ndarray
        Array of coordinates of reference locations in
        (m, n-dimensional) order, such as {n_locations,
        x_coordinates, y_coordinates, ...} for as many dimensions as
        required. For example to calculate distances to a single x,y,z
        location:

        refs = [-57.345, -110.134, 1012]

        And to calculate distances to three x,y reference locations:

        refs = [
            [-57.345, -110.134],
            [-56.345, -109.123],
            [-58.534, -112.123]
        ]

        The supplied array has to have at least x,y coordinates with a
        (1, 2) shape for a single location.

    minimum : bool, default is False
        Optionally calculate the minimum distance to the combined
        reference locations, resulting in a single new feature,
        rather than a new feature for each individual reference
        location.

    log : bool (opt), default=False
        Optionally log-transform the distance measures.

    Returns
    -------
    X_new : ndarray
        Array of shape (n_samples, n_features) with new geodistance
        features appended to the right-most columns of the array.
    """

    def __init__(self, refs, minimum=False, log=False):
        self.refs = refs
        self.log = log
        self.refs_ = None
        self.minimum = minimum

    def fit(self, X, y=None):
        self.refs_ = np.asarray(self.refs)

        if self.refs_.ndim < 2:
            raise ValueError(
                "`refs` has to be a m,n-dimensional array with  at least two dimensions"
            )

        return self

    def transform(self, X, y=None):
        if self.minimum is False:
            dists = cdist(self.refs_, X).transpose()

        if self.minimum is True:
            tree = cKDTree(self.refs_)
            dists, _ = tree.query(X)

        if self.log is True:
            dists = np.log(dists)

        return np.column_stack((X, dists))


def _apply_transformer(img, transformer):
    img = np.ma.masked_invalid(img)
    mask = img.mask.copy()

    # reshape into 2D array
    n_features, rows, cols = img.shape[0], img.shape[1], img.shape[2]
    flat_pixels = img.reshape((rows * cols, n_features))
    flat_pixels = flat_pixels.filled(0)

    # predict and replace mask
    result = transformer.transform(flat_pixels)

    # reshape the prediction from a 1D into 3D array [band, row, col]
    result = result.reshape((n_features, rows, cols))
    result = np.ma.masked_array(data=result, mask=mask, copy=True)
    
    return result



from unittest import TestCase
from tempfile import NamedTemporaryFile

from pyspatialml import Raster
from pyspatialml.datasets import nc


class TestWrite(TestCase):
    def setUp(self) -> None:
        # inputs
        self.predictors = [nc.band1, nc.band2, nc.band3, nc.band4, nc.band5,
                           nc.band7]

        # test results
        self.stack = None

    def tearDown(self) -> None:
        self.stack.close()

    def test_write(self):
        # test writing to file
        self.stack = Raster(self.predictors)
        fp = NamedTemporaryFile(suffix=".tif").name

        result = self.stack.write(fp)

        self.assertIsInstance(result, Raster)
        self.assertEqual(result.count, self.stack.count)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    # package metadata
    name="pyspatialml",
    version="0.2",
    author="Steven Pawley",
    author_email="steven.pawley@gmail.com",
    description=("Machine learning for GIS spatial modelling"),
    long_description=README,
    long_description_content_type="text/markdown",
    license="GNU",
    keywords="GIS",
    url="https://github.com/stevenpawley/pyspatialml",

    # files/directories to be installed with package
    packages=find_packages(),
    package_data={
        '': ['*.tif', '*.dbf', '*.prj', '*.shp', '*.shx'],
    },
    include_package_data=True,

    # package dependencies
    install_requires=[
        'numpy>=1.10.0',
        'scipy>1.0.0',
        'tqdm>=4.20',
        'rasterio>=1.0',
        'pandas>=0.20',
        'shapely>=1.6',
        'geopandas>=0.3',
        'matplotlib>=2.2.4',
        'scikit-learn>=0.22'],
    python_requires='>=3.6',

    # testing
    setup_requires=['pytest-runner'],
    tests_require=['pytest']
)
